;;; sh-beg-end.el --- Something for C-M-a,
;;; C-M-e, M-a and M-e in shell-script-mode

;; Version 1.4

;; Copyright (C) 2007 by Andreas Rhler
;; <andreas.roehler@online.de>

;; Keywords: languages

;; This file is free software; you can redistribute it
;; and/or modify it under the terms of the GNU General
;; Public License as published by the Free Software
;; Foundation; either version 3, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the
;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;; PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.

;; You should have received a copy of the GNU General
;; Public License along with GNU Emacs; see the file
;; COPYING.  If not, write to the Free Software
;; Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.




;;; Commentary:

;; C-M-a, C-M-e: jump to the beginning or end of a
;; top-level-form in sh-mode -
;; "Shell-script"-mode.  Repeated execution forward
;; or backward.  With argument as many times.

;; M-a, M-e: jump to the beginning or end of command in
;; a given line, forward or backward next beginning or
;; end.  With argument do this as many times.

;; Thanks to Stefan Monnier and Richard Stallman
;; being helpful.  Errors are mine, any comments
;; welcome.

;;; Code:

(require 'sh-script)
;; Some suggestions for keys
(define-key sh-mode-map "\M-a" 'sh-beg-of-command)
(define-key sh-mode-map "\M-e" 'sh-end-of-command)
(define-key sh-mode-map "\C-\M-a" 'sh-beginning-of-form)
(define-key sh-mode-map "\C-\M-e" 'sh-end-of-form)
(define-key sh-mode-map "\M-p" 'sh-goto-opening)
(define-key sh-mode-map "\M-n" 'sh-goto-closing)
(define-key sh-mode-map "%" 'sh-match-paren)

(setq beg-sh-struct-atpt "\\bcase\\b\\|\\bfor\\b\\|\\bfunction\\b\\|\\bif\\b\\|\\bselect\\b\\|\\buntil\\b\\|\\bwhile\\b")

(setq end-sh-struct-atpt "\\bdone\\b\\|\\besac\\b\\|\\bfi\\b") 

(setq sh-match-paren-char "%")

(defcustom sh-beginning-of-form-regexp "^[A-Za-z_][A-Za-z_0-9]*"
  "Regexp indicating the beginning of a form to edit."
  :type 'regexp
  :group 'lisp)

(defcustom sh-end-of-form-regexp "^}[ \t]*$\\|;}[ \t]*$"
  "Regexp indicating the end of a form to edit."
  :type 'regexp
  :group 'lisp)

(defun sh-beginning-of-form (&optional arg)
  "Move to the beginning of a top-level-form in sh-script.
With numeric ARG, do it that many times."
  (interactive "p")
  (let ((arg (or arg 1)))
    (re-search-backward sh-beginning-of-form-regexp nil t arg)))

;;  Lists

;; (put 'list 'end-op-atpt (lambda () (search-forward "fi" nil t 1)))
;; (put 'list 'beginning-op-atpt (lambda () (search-backward "if" nil t 1)))

(defun sh-end-of-form (&optional arg)
  "Move forward to end of a top-level-form or command in sh-script.
With numeric ARG, do it that many times."
  (interactive "p")
  (let ((arg (or arg 1)))
    (when (re-search-forward sh-end-of-form-regexp nil 'move arg)
      (skip-chars-backward " \t\r\n\f"))))
      
(defun sh-set-beginning-of-form ()
  "Set `sh-beginning-of-form' as `beginning-of-defun-function'."
  (interactive)
  (set (make-local-variable 'beginning-of-defun-function) 'sh-beginning-of-form))

(defun sh-set-end-of-form ()
  "Set `sh-end-of-form' as `end-of-defun-function'."
  (interactive)
  (set (make-local-variable 'end-of-defun-function) 'sh-end-of-form))

(defun sh-beg-of-command (&optional arg)
  "Move point to ARG successive beginnings of commands."
  (interactive "p")
  (let ((arg (or arg 1))
	(pos (point)))
    (back-to-indentation)
    (unless (eq pos (point))
      (setq arg (1- arg)))
    (while (< 0 arg)
      (forward-line (- arg))
      (setq arg (1- arg))
      ;; skip comments and empty lines and closing braces
      (let ((pos (point)))
	(if (forward-comment -1)
	    (while (forward-comment -1) (forward-comment -1))
	  (goto-char pos)))
      (while (or (empty-line-p)
		 (looking-at "}"))
	(forward-line -1))
      (back-to-indentation))))

(defun sh-end-of-command (&optional arg)
  "Move point to ARG successive ends of commands."
  (interactive "p")
  (let ((arg (or arg 1))
	(pos (point)))
    (end-of-line)
    (skip-chars-backward " \t\r\n\f" (line-beginning-position))
    (unless (eq pos (point))
      (setq arg (1- arg)))
    (while (< 0 arg)
      (forward-line arg)
      (setq arg (1- arg)))
    (end-of-line)
    (skip-chars-backward " \t\r\n\f" (line-beginning-position))
    (while (or
	    (empty-line-p)
	    (forward-comment 1))
      (forward-line 1)
      (end-of-line))
    (skip-chars-backward " \t\r\n\f")))

(defun sh-mark-function ()
  " "
  (interactive)
  (sh-beginning-of-form)
  (push-mark (point) t t)
  (sh-end-of-form)
  (kill-new (buffer-substring-no-properties (mark) (point))))

(defun sh-kill-function ()
  " "
  (interactive)
  (sh-beginning-of-form)
  (push-mark (point) t t)
  (sh-end-of-form)
  (kill-region (mark) (point)))

(defun sh-transpose-funktion ()
  " "
  (interactive "*")
  (re-search-backward "(.*)" nil t 1)
  (sh-kill-function)
  (re-search-forward "(.*)" nil t 1)
  (sh-end-of-form)
  (newline 2)
  (insert (car kill-ring)))

(defcustom empty-line-p-chars "^[ \t\f\r]*$"
  "Empty-line-p-chars."
 :type 'regexp
 :group 'convenience)

(defun empty-line-p (&optional ispec)
  "Returns t if cursor is at an empty line, nil otherwise."
  (interactive "p")
  (save-excursion
    (beginning-of-line)
    (when ispec
      (message "%s" (looking-at empty-line-p-chars)))
    (looking-at empty-line-p-chars)))

(defun sh-opening-closing ()
  "Switch between opening and closing programming constructs.
If not at the beginning or the end, to opening first. "
  (interactive)
  (cond ((looking-at beg-sh-struct-atpt)
	 (sh-goto-closing))
	((progn (backward-word)
		(looking-at end-sh-struct-atpt))
	 (sh-goto-opening))
	(t (sh-goto-opening))))

(defun sh-goto-opening ()
  "Goto opening of a programming structure in this level. "
  (interactive)
  (let ((pos (point))) 
    (if (and (not (looking-at "[ \t\f\n\r]")) (not (looking-back "[ \t\f\n\r]")))
	(progn 
	  (backward-word)
	  (when (< (point) pos)
	    ;; point moved, now at the beginning
	    (cond ((looking-at beg-sh-struct-atpt)
		   nil)
		  ((looking-at end-sh-struct-atpt)
		   (sh-goto-opening-intern))
		  (t (sh-goto-opening-intern)))))
      (sh-goto-opening-intern))))
	    
(defun sh-goto-opening-intern () 
  (let ((nesting 0)) 
    (while
	(and
	 (<= 0 nesting))
      (if (and (re-search-backward (concat "\\(" beg-sh-struct-atpt "\\)\\|\\(" end-sh-struct-atpt "\\)") nil t 1) (string-match end-sh-struct-atpt (match-string-no-properties 0)))
	  (setq nesting (1+ nesting))
	(setq nesting (1- nesting))))))

(defun sh-goto-closing ()
  "Goto closing of a programming structure in this level. "
  (interactive)
  (when (looking-at beg-sh-struct-atpt)
    (forward-word 1)) 
  (let ((nesting 0)) 
    (while
	(and
	 (<= 0 nesting))
      (re-search-forward (concat "\\(" beg-sh-struct-atpt "\\)\\|\\(" end-sh-struct-atpt "\\)") nil t 1)
      (if
	  (string-match beg-sh-struct-atpt (match-string-no-properties 0))
	  (setq nesting (1+ nesting))
	(setq nesting (1- nesting)))))
  (forward-char -1))

(defun sh-match-paren (arg &optional ins) 
  "Go to the matching opening/closing. 
First to opening, unless cursor is already there.
With arg, insert the charakter of `sh-match-paren-char'.
Key per default is \"%\" as with elisp's `match-paren'. "
  (interactive "p\nP")
  (if ins
      (insert sh-match-paren-key-char)
    (cond ((looking-at beg-sh-struct-atpt)
	   (sh-goto-closing))
	  ((looking-at "\\s(")
	   (match-paren arg))
	  ((looking-at "\\s)")
	   (match-paren arg))
	  (t (sh-goto-opening)))))

(defvar sh-match-paren-mode nil)

(defun sh-match-paren-mode (&optional ispec) 
  "Toggle sh-match-paren-mode. 
If on, inserting of `sh-match-paren-char', default is \"%\", moves to the matching opening/closing.
With arg, insert the charakter the key is on 
Key per default is \"%\" as with elisp's `match-paren'. "
  (interactive "p")
  (if sh-match-paren-mode
      (progn 
	(setq sh-match-paren-mode nil)
	(define-key sh-mode-map "%" 'self-insert-command)
	(when ispec (message "sh-match-paren-mode: %s" sh-match-paren-mode)))
    (setq sh-match-paren-mode t)
    (define-key sh-mode-map "%" 'sh-match-paren)
    (when ispec (message "sh-match-paren-mode: %s" sh-match-paren-mode))))

(defconst sh-beg-end-version "1.4"
  "Version number of this sh-beg-end.el.")

(defun sh-beg-end-version (&optional ispec)
  "Print and/or return sh-beg-end-version."
  (interactive "p")
  (when ispec
    (message "%s" sh-beg-end-version))
  sh-beg-end-version)

(add-hook 'sh-mode-hook 'sh-set-beginning-of-form)
(add-hook 'sh-mode-hook 'sh-set-end-of-form)

(provide 'sh-beg-end)

;;; sh-beg-end.el ends here
