
;;; thingatpt-utils-base.el --- th-at-point edit functions

;; Version: 1.10

;; Copyright (C) 2006, 2007, 2008 Andreas Roehler

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;; Keywords: convenience

;;; Major changes to previous version: 

;; New functions `functionname-', `ip-'.  
;; Chars which are accepted as part of a URL are
;; delivered by a var `url-at-point-chars', which is
;; customizable. '()' now are accepted as part of a URL
;; right out of the box.

;;; Commentary:

;; A set of functions to return, mover over or
;; manipulate a given THING. THING may be a well known
;; form as symbol', `list', `sexp', `defun' but also a
;; new defined and abstract thing. 

;; Call one of the test-functions `th-mv-test',
;; `th-delimtest', `th-test' with come chars in
;; scratch-buffer or any else changable buffer to get
;; an impression

;; The idea comes from Mike Williams
;; <mikew@gopher.dosli.govt.nz>, author of
;; thingatpt.el

;; The goal is to have a set of similar forms. For
;; example, to provide a word with double-quotes around
;; it, call doublequote-word-at. In a similar way you
;; may double-quote not just a word, but any object
;; instrumented here as THING. To make parentheses
;; around it call parentize-word-at, etc.

;; Move-functions of this package differ from common
;; behaviour in such, as `forward-word' i.e. stops
;; _not_ _after_ THING, but on the last char of
;; THING. That's in order to enable a call of
;; thing-at-point functions at the end
;; position. Otherwise, when cursor stops after word
;; (THING) finally, word-atpt would return nil. So it
;; returns word, boundaries of word or whatever is
;; called (and implemented).

;; To see other features, maybe try separate-list-at
;; or comment-list-atpt while point is inside a
;; list. Try it again with an abstract char-class as
;; [:alnum:], i.e. try comment-alnum-at,
;; brace-alnum-atpt etc.

;; Call `list-of-things-at' to see which objects are
;; presently instrumented, beside `delimited-at', where
;; selection specified on the fly (because of the wide range of
;; chars, designing such a possible THING).

;; Most THINGS are provided with a set of functions at,
;; before and after point - i.e. to call with
;; ACTION-THING-at, -bf, af.

;; This utility comes with test-functions which return
;; the possible results of most functions (exception
;; are the kill-fns). Call th-test, th-mv-test
;; or th-delimtest over text. That-delimtest
;; changes but restores the buffer. Customize the speed
;; of execution via `th-delimtest-delay' and
;; `th-mv-test-delay.'

;; Diffs to basics of required thingatpt.el:
;; `bounds-of-thing-at-point' is replaced by a new
;; `th-bounds-intern', which now first searches
;; backward. As a consequence several
;; `beginning-op-at' and `end-op-at' constructs had
;; to be rewritten.

;; Behavior in general is not validating; i.e. if you
;; call url-atpt and there is no url, all chars at
;; point may be picked, which could be part of a
;; url. Sometimes, however, a kind of validation may be
;; introduced.

;; In case of trouble, please send me a bug report. Any
;; ideas and comments welcome.

;; How it works:

;; Thing-at-point delivers a portion of the
;; buffer. This substring is determined by
;; move-functions specified for thingatpt, called
;; beginning-op-at and end-op-at. Point is stored
;; after move, beginning and end delivered as pair: as
;; consed bounds-of-thing. It's easy to write your own
;; thing-at-point functions that way. You need the
;; caller and both move forms:

;; (defun MY-FORM-atpt (&optional arg ispec) 
;;   " "
;;   (interactive "p\np")
;;   (th 'MY-FORM arg ispec))

;; (put 'MY-FORM 'beginning-op-at 
;;            (lambda () MY-FORWARD-MOVE-CODE))

;; (put 'MY-FORM 'end-op-at
;;      (lambda () MY-BACKWARD-MOVE-CODE))

;; For example if you want to pick all chars at point
;; which are written between a string "AAA" and a
;; "BBB", which may exist as
;; AAA Luckily detected a lot of things! BBB
;; After evaluation of
;; (put 'MY-FORM 'beginning-op-at
;;      (lambda ()
;;        (search-backward "AAA" nil t 1)
;;        ;; step chars of search expression back
;;        (forward-char 3)))
;; 
;; (put 'MY-FORM 'end-op-at
;;      (lambda ()
;;        (search-forward "BBB" nil t 1)
;;        (forward-char -3)))
;; together with the functions definition above, it's ready.
;; M-x MY-FORM-atpt 
;; (while point inside) you should see: 
;; " Luckily detected a lot of things! " 
;; in the minibuffer.

;;; Code:

(require 'string-strip)
(require 'sh-beg-end)

;; (defvar th-orig 0
;; "Correct orig according to delimiter-length")

(if (featurep 'xemacs) 
    (setq thatpt-directory "~/xemacs-thingatpt")
  (setq thatpt-directory "~/emacs-thingatpt"))

(make-directory thatpt-directory t)

;; Ascii

(unless (featurep 'xemacs)
 (put 'ascii 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:ascii:]]")
       (skip-chars-backward "[:ascii:]")))))

(unless (featurep 'xemacs)
 (put 'ascii 'end-op-at
     (lambda ()
       (skip-chars-forward "[:ascii:]"))))


;; Alnum

(unless (featurep 'xemacs)
 (put 'alnum 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:alnum:]]")
       (skip-chars-backward "[:alnum:]")))))

(unless (featurep 'xemacs)
 (put 'alnum 'end-op-at
     (lambda ()
       (skip-chars-forward "[:alnum:]"))))


;; Alpha

(unless (featurep 'xemacs)
 (put 'alpha 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:alpha:]]")
       (skip-chars-backward "[:alpha:]")))))

(unless (featurep 'xemacs)
 (put 'alpha 'end-op-at
     (lambda ()
       (skip-chars-forward "[:alpha:]"))))


;; Blank

(unless (featurep 'xemacs)
 (put 'blank 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:blank:]]")
       (skip-chars-backward "[:blank:]")))))

(unless (featurep 'xemacs)
 (put 'blank 'end-op-at
     (lambda ()
       (skip-chars-forward "[:blank:]"))))


;; Cntrl

(unless (featurep 'xemacs)
 (put 'cntrl 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:cntrl:]]")
       (skip-chars-backward "[:cntrl:]")))))

(unless (featurep 'xemacs)
 (put 'cntrl 'end-op-at
     (lambda ()
       (skip-chars-forward "[:cntrl:]"))))


;; Digit

(unless (featurep 'xemacs)
 (put 'digit 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:digit:]]")
       (skip-chars-backward "[:digit:]")))))

(unless (featurep 'xemacs)
 (put 'digit 'end-op-at
     (lambda ()
       (skip-chars-forward "[:digit:]"))))


;; Graph

(unless (featurep 'xemacs)
 (put 'graph 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:graph:]]")
       (skip-chars-backward "[:graph:]")))))

(unless (featurep 'xemacs)
 (put 'graph 'end-op-at
     (lambda ()
       (skip-chars-forward "[:graph:]"))))


;; Lower

(unless (featurep 'xemacs)
 (put 'lower 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:lower:]]")
       (skip-chars-backward "[:lower:]")))))

(unless (featurep 'xemacs)
 (put 'lower 'end-op-at
     (lambda ()
       (skip-chars-forward "[:lower:]"))))


;; Multibyte

(unless (featurep 'xemacs)
 (put 'multibyte 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:multibyte:]]")
       (skip-chars-backward "[:multibyte:]")))))

(unless (featurep 'xemacs)
 (put 'multibyte 'end-op-at
     (lambda ()
       (skip-chars-forward "[:multibyte:]"))))


;; Nonascii

(unless (featurep 'xemacs)
 (put 'nonascii 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:nonascii:]]")
       (skip-chars-backward "[:nonascii:]")))))

(unless (featurep 'xemacs)
 (put 'nonascii 'end-op-at
     (lambda ()
       (skip-chars-forward "[:nonascii:]"))))


;; Print

(unless (featurep 'xemacs)
 (put 'print 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:print:]]")
       (skip-chars-backward "[:print:]")))))

(unless (featurep 'xemacs)
 (put 'print 'end-op-at
     (lambda ()
       (skip-chars-forward "[:print:]"))))


;; Punct

(unless (featurep 'xemacs)
 (put 'punct 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:punct:]]")
       (skip-chars-backward "[:punct:]")))))

(unless (featurep 'xemacs)
 (put 'punct 'end-op-at
     (lambda ()
       (skip-chars-forward "[:punct:]"))))


;; Space

(unless (featurep 'xemacs)
 (put 'space 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:space:]]")
       (skip-chars-backward "[:space:]")))))

(unless (featurep 'xemacs)
 (put 'space 'end-op-at
     (lambda ()
       (skip-chars-forward "[:space:]"))))


;; Unibyte

(unless (featurep 'xemacs)
 (put 'unibyte 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:unibyte:]]")
       (skip-chars-backward "[:unibyte:]")))))

(unless (featurep 'xemacs)
 (put 'unibyte 'end-op-at
     (lambda ()
       (skip-chars-forward "[:unibyte:]"))))


;; Upper

(unless (featurep 'xemacs)
 (put 'upper 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:upper:]]")
       (skip-chars-backward "[:upper:]")))))

(unless (featurep 'xemacs)
 (put 'upper 'end-op-at
     (lambda ()
       (skip-chars-forward "[:upper:]"))))


;; Xdigit

(unless (featurep 'xemacs)
 (put 'xdigit 'beginning-op-at
     (lambda ()
       (when
         (looking-at "[[:xdigit:]]")
       (skip-chars-backward "[:xdigit:]")))))

(unless (featurep 'xemacs)
 (put 'xdigit 'end-op-at
     (lambda ()
       (skip-chars-forward "[:xdigit:]"))))



(defconst th-version "1.10" "\
Version number of this version of thingatpt-utils.")

(defun th-version (&optional ispec) 
  "Print and/or return th-version"
  (interactive "p")
  (when ispec
    (message "%s" th-version))
  th-version)

(defvar thingatpt-listed nil
  "Forms listed with thingatpt-list")

(setq thingatpt-listed '(alnum alpha ascii backslashed blank braced bracketed cntrl comment csv date defun delimited digit dollared doublequoted email filename float function graph ip isbn line list lower ml-text multibyte nonascii number page paragraph paren parentized phone print punct region sentence sexp singlequoted slashed space string sh-struct symbol unibyte upper url whitespace word xdigit))

(defun thingatpt-listed ()
  (interactive "*")
  (message "%s" thingatpt-listed))


;; enhance compatibility with XEmacs
(unless (featurep 'xemacs) 
  (defsubst region-exists-p ()
    "(not (null (mark)))"
    (not (null (mark)))))

(unless (featurep 'xemacs)
  (defun region-active-p ()
    "(and mark-active transient-mark-mode
	   (not (eq (region-beginning) (region-end))))"
    (and mark-active transient-mark-mode
         (not (eq (region-beginning) (region-end))))))

;; opposite proceeding: from GNU Emacs subr.el
(when (featurep 'xemacs) 
  (defun looking-back (regexp &optional limit greedy)
    "Return non-nil if text before point matches regular expression REGEXP.
    Like `looking-at' except matches before point, and is slower.
    LIMIT if non-nil speeds up the search by specifying a minimum
    starting position, to avoid checking matches that would start
    before LIMIT.
    If GREEDY is non-nil, extend the match backwards as far as possible,
    stopping when a single additional previous character cannot be part
    of a match for REGEXP."
    (let ((start (point))
	  (pos
	   (save-excursion
	     (and (re-search-backward (concat "\\(?:" regexp "\\)\\=") limit t)
		  (point)))))
      (if (and greedy pos)
	  (save-restriction
	    (narrow-to-region (point-min) start)
	    (while (and (> pos (point-min))
			(save-excursion
			  (goto-char pos)
			  (backward-char 1)
			  (looking-at (concat "\\(?:"  regexp "\\)\\'"))))
	      (setq pos (1- pos)))
	    (save-excursion
	      (goto-char pos)
	      (looking-at (concat "\\(?:"  regexp "\\)\\'")))))
      (not (null pos)))))


(defun face-atpt (&optional ispec) 
  "Return face at point "
  (interactive "p")
  (let ((faceat (get-char-property (point) 'face)))
    (when ispec (message "%s" faceat))
    faceat))

(defun customize-face-atpt ()
  "Make customization buffer for face at point. "
  (interactive)
  (customize-variable (cadr (text-properties-at (point) (current-buffer)))))

;; Braced

(put 'braced 'beginning-op-at
     (lambda ()
       (if (looking-at "{")
	   (point)
	 (let* ((end (point))
		(beg (progn (beginning-of-defun) (point)))
		(rekord (parse-partial-sexp beg end))
		(pos (cadr rekord))
		(level (car rekord)))
	   ;; (with-syntax-table sh-mode-default-syntax-table (parse-partial-sexp (point-min) (point) 0))
	   ;; (while (eq level (with-syntax-table (sh-mode-syntax-table (car (parse-partial-sexp (point-min) (point) 0)))))
	   ;; (forward-char 1)))))
	   (goto-char pos)
	   (unless (looking-at "{")
	     (unless (search-backward "{" beg t 1)
	       (when (search-forward "{" end nil 1)
		 (forward-char -1))))))))

(put 'braced 'end-op-at
     (lambda ()
       (forward-list 1)
       (unless (looking-back "}")
	 (search-forward "}" nil nil 1))))

;; Bracketed

(put 'bracketed 'beginning-op-at
     (lambda ()
       (if (looking-at "[[]")
	   (point)
	 (let* ((pos (point))
                (beg (search-backward "[" nil t 1)))))))
 		
(put 'bracketed 'end-op-at
     (lambda ()
       (search-forward "]" nil t 1)))

;;; CSV

;; Inspired by
;;; csv-mode.el --- major mode for editing comma-separated value files
;; Author: Francis J. Wright <F.J.Wright at qmul.ac.uk>
;; URL: http://centaur.maths.qmul.ac.uk/Emacs/

(defcustom csv-separator-atpt ";"
  "Char to distinguish datasets in a `comma`-separated row" 
  :type 'string
  :group 'convenience)

;; (when (boundp 'csv-separators)
;; (setq separator-atpt csv-separators))

(put 'csv 'beginning-op-at
     (lambda ()
       (search-backward (concat "^" csv-separator-atpt) nil nil 1)))

(put 'csv 'end-op-at
     (lambda ()
       (skip-chars-forward (concat "^" csv-separator-atpt))))

;; DATE

(put 'date 'beginning-op-at
     (lambda ()
       ;; provide for the case, we are over a
       ;; string-delimiter as `"'
       (when
	   (and (not (eq 32 (char-after))) 
		(or (bobp)
		    (eq 32 (char-before))))
	 (forward-char 1)
	 ;; as the bounds-function checks position, correct it
	 ;; (setq th-orig 1)
	 ) 
       (skip-chars-backward "[0-9] .-")
       (skip-chars-forward " ") 
))

(put 'date 'end-op-at
     (lambda ()
              (skip-chars-forward "[0-9] .-")
	      (skip-chars-backward " ")))

;; Defun

(put 'defun 'beginning-op-at (lambda (&optional arg) (beginning-of-defun (or arg 1))))

(put 'defun 'end-op-at (lambda (&optional arg)(end-of-defun (or arg 1))))

;; Delimited


(put 'delimited 'beginning-op-at
     (lambda ()
       (unless delimiters-at
	   (setq delimiters-at "\"'#$/=?!:;"))
       (if
	   (looking-at (concat "[^" delimiters-at "]"))
	   (progn 
	     (re-search-backward (concat "[" delimiters-at "]") nil nil 1)
	     (setq delimiter-zeichen-atpt (char-after)))
	 (setq delimiter-zeichen-atpt nil)))) 

(put 'delimited 'end-op-at
     (lambda ()
       (forward-char 1)
     (search-forward (char-to-string delimiter-zeichen-atpt) nil nil 1)))

;; (put 'delimited 'beginning-op-at
;;      (lambda ()
;;        (when
;; 	   (looking-at (concat "[^" delimiters-at "]"))
;; 	 (re-search-backward (concat "[" delimiters-at "]") nil nil 1)
;; 	 (setq delimiter-zeichen-atpt (char-after))  
;; 	   (when with-delimiters-at
;; 	     (forward-char -1))
;; t)))
;; 
;; (put 'delimited 'end-op-at
;;      (lambda ()
;;        (when with-delimiters-at
;; 	 (forward-char 2))
;;        (re-search-forward (char-to-string delimiter-zeichen-atpt) nil nil 1)))


(defvar delimiter-zeichen-atpt nil
  "Delimiter char found at place, search it backward then")

;; (defvar delimiters-at nil
;; "Accepted delimiters chars")

(defcustom with-delimiters-at  nil 
 "If containing delimiters are copied with string" 

:type 'boolean
:group 'convenience
)

(defcustom delimiters-at "\"'#$/=?!:;"
  "Specify the accepted delimiter chars."
  :type 'string
  :group 'convenience)

(defcustom th-beg-delimiter "{[(/"
  "Specify the delimiter char."
  :type 'string
  :group 'convenience)
				
(defcustom th-end-delimiter "}])/"
  "Specify the delimiter char."
  :type 'string
  :group 'convenience)

;; (defun set-delimiter-atpt (&optional delimiters-atpt)
;; "Enables interactive setting of delimiters. "
;; (if delimiters-at
;; (setq th-beg-delimiter (car delimiters-atpt) th-end-delimiter (cdr delimiters-atpt))
;; (setq th-beg-delimiter
;; (read-from-minibuffer "th-beg-delimiter: " th-beg-delimiter))
;; (setq th-end-delimiter
;; (read-from-minibuffer "th-end-delimiter: " th-end-delimiter))))

;; Doublequoted


(put 'doublequoted 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^\"]"))
       ;; beg or end of Doublequoted?
       (unless (looking-back "[\t\r\n\f ]") 
	 (search-backward "\"" nil nil 1))))

(put 'doublequoted 'end-op-at
     (lambda ()
       (when (looking-at "\"") 
	 (search-forward "\"" nil nil 2))))

;; Email

(put 'email 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^ \t]")
	 (re-search-backward "[,;][[:graph:]]\\|<[[:graph:]]\\|^[[:graph:]]\\|[^[:graph:]][[:graph:]]" (line-beginning-position) t 1)(when (looking-at "[[:space:];,]") (forward-char 1))))) 

(put 'email 'end-op-at (lambda () (when (looking-at "[ <]\\{0,1\\}\\([\041-\132\136-\176]+@[\041-\132\136-\176]+\\)[;,> \t\n]*")
				 (goto-char (match-end 1))
				 (skip-chars-backward "[[:punct:]]")))) 

			   ;; (put 'email 'end-op-at (lambda () (when (looking-at "[ <]\\{0,1\\}\\([[:graph:]]+@[[:graph:]]+\\)[;,> \t\n]*")
			   ;; 			   (goto-char (match-end 1))
			   ;; 			   (skip-chars-backward "[[:punct:]]"))))

;; Filename

; (setq thingatpt-file-name-chars "-~/A-Za-z0-9ÄÖÜäöüß_.$?={}#%,: ") 
(if (featurep 'xemacs)
    (defcustom thingatpt-file-name-chars "-~/A-Za-z0-9ÄÖÜäöüß_.$?={}#%,: " 
      "Characters forseen in filenames. "
      :type 'string
      :group 'convenience)
  (defcustom thingatpt-file-name-chars "-~/[:alnum:]_.$?={}#%,:" 
    "Characters forseen in filenames. "
    :type 'string
    :group 'convenience)) 

; (defcustom thingatpt-file-name-chars  "-~/[:alnum:]_.$?={}#%,:" 
; "Characters forseen in filenames. "
; :type 'string
; :group 'convenience)

(put 'filename 'beginning-op-at
     (lambda ()
       (unless (looking-back " \t\n\r\f")
	 (skip-chars-backward thingatpt-file-name-chars))))

(put 'filename 'end-op-at
     (lambda ()
       (re-search-forward (concat "\\=[" thingatpt-file-name-chars "]*")
			  nil t)
     (skip-chars-backward ": ")))

;; Floats

(put 'float 'beginning-op-at (lambda () (when (numberp (read (buffer-substring-no-properties (point) (1+ (point)))))
    (skip-chars-backward "[0-9].,"))))

(put 'float 'end-op-at (lambda () (skip-chars-forward "[0-9.,]")))

;; Function

(put 'function 'beginning-op-at
     (lambda ()
       (unless (or
		(and (eq major-mode 'sh-mode) (looking-at sh-beginning-of-form-regexp))
		(and (eq 4 (car (syntax-after (point)))) (eq 0 (car (parse-partial-sexp (point-min) (point))))))
	 (beginning-of-defun)))) 

(put 'function 'end-op-at
     (lambda ()
       (end-of-defun)
       (when (string= major-mode "emacs-lisp-mode")
	 (skip-chars-backward " \t\r\n\f")))) 

;; IP

(put 'ip 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-chars-backward "[0-9]."))))

(put 'ip 'end-op-at
     (lambda ()
       (when (looking-at "[0-9]\\{1,3\\}.[0-9-]\\{1,3\\}.[0-9]\\{1,3\\}.[0-9]\\{1,3\\}")
	 (goto-char (match-end 0)))))

;; ISBN

(put 'isbn 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-chars-backward "[0-9]-"))))

(put 'isbn 'end-op-at
     (lambda ()
       (when (looking-at "[0-9]\\{1,3\\}[0-9-]\\{7,12\\}[0-9X]\\{0,1\\}")
	 (goto-char (match-end 0)))))

;; Lines

(put 'line 'beginning-op-at (lambda () (beginning-of-line)))
(put 'line 'end-op-at (lambda () (end-of-line)))
   
;;  Lists  


(put 'list 'end-op-at (lambda () (forward-list 1)
)) 

(put 'list 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s(")
	 (if (nth 9 (parse-partial-sexp (defun-beginning-position) (point)))
	     (goto-char (car (last (nth 9 (parse-partial-sexp (defun-beginning-position) (point))))))
	   (backward-sexp)
	   (forward-char -1)))))

;; Ml-Text
;; Useful to extract texts between ml-tags

(put 'ml-text 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^>]") 
	 (re-search-backward ">" nil nil 1)
         (forward-char 1)))) 

(put 'ml-text 'end-op-at (lambda () (re-search-forward "</" nil nil 1) (forward-char -2))) 

;; Number

(put 'number 'beginning-op-at
     (lambda ()
       (when
	   (numberp (string-to-number (buffer-substring-no-properties (point) (1+ (point)))))
	 (skip-chars-backward "[0-9]"))))

(put 'number 'end-op-at
 (lambda ()
   (skip-chars-forward "[0-9]")))


;; Page

(put 'page 'beginning-op-at
     (lambda ()
       (backward-page)))


(put 'page 'end-op-at
     (lambda ()
       (forward-page)))


;; Paragraph

(put 'paragraph 'beginning-op-at
     (lambda ()
       (backward-paragraph)))


(put 'paragraph 'end-op-at
     (lambda ()
       (forward-paragraph)))

;; Paren
(put 'paren 'beginning-op-atpt
     (lambda ()
       (cond
	((looking-at "\\s)")
	 (forward-char 1) (backward-list 1))
	(t (while
	       (and (not (looking-at "\\s(")) (not (bobp)))
	     (forward-char -1))))))

(put 'paren 'end-op-atpt
     (lambda ()
       (forward-list 1)))

;; Parentized

(put 'parentized 'beginning-op-at
     (lambda ()
       (if (looking-at "(")
	   (point)
	 (let* ((end (point))
		 (beg (progn (beginning-of-defun) (point)))
		 (rekord (parse-partial-sexp beg end))
		 (pos (cadr rekord))
		 (level (car rekord)))
		(goto-char pos)
		(unless (looking-at "(")
		  (unless (search-backward "(" beg t 1)
		    (when (search-forward "(" end nil 1)
		      (forward-char -1))))))))

(put 'parentized 'end-op-at
     (lambda ()
       (forward-list 1)
       (unless (looking-back ")")
	 (search-forward ")" nil nil 1))))

;; Phone

(put 'phone 'beginning-op-at
     (lambda ()
       (when
	   (and (looking-at "[0-9 \t.()-]")
		(not (eq (char-before) ?+)))
       (re-search-backward "[^0-9 \t.()-][0-9 ()\t-]+" (line-beginning-position) nil 1) (forward-char 1)))) 

(put 'phone 'end-op-at
     (lambda ()
       (when
	   (looking-at "[0-9;, \t()-]")
	 (re-search-forward "[0-9 \t.()-]+[^0-9 \t-]" (1+ (line-end-position)) nil 1) (forward-char -1)))) 

;; Region

(defvar region-end-atpt nil)

(put 'region 'end-op-at
     (lambda ()
       (goto-char region-end-atpt)))

(put 'region 'beginning-op-at
     (lambda ()
       (setq region-end-atpt (region-end))
       (goto-char (region-beginning))))


;; Sentence

(put 'sentence 'beginning-op-at
     (lambda ()
       (backward-sentence)))


(put 'sentence 'end-op-at
     (lambda ()
       (forward-sentence)))


;; Sexp

(put 'sexp 'beginning-op-at
     (lambda ()
       (backward-sexp)))


(put 'sexp 'end-op-at
     (lambda ()
       (forward-sexp)))

;; Singlequoted

(put 'singlequoted 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^']")
	 (search-backward  "'" nil nil 1))))


(put 'singlequoted 'end-op-at
     (lambda ()
       (search-forward "'" nil nil 2)))

;; Slashed

(put 'slashed 'beginning-op-at
     (lambda ()
       (if (looking-at "/")
	   (point)
	 (search-backward "/" nil nil 1))))

(put 'slashed 'end-op-at
     (lambda ()
       (search-forward "/" nil nil 2)))

;; Strings

(defvar th-string-beginningp nil)

(put 'string 'beginning-op-at
     (lambda ()
       (setq th-string-beginningp nil)
       (let ((pos (with-syntax-table (standard-syntax-table) (nth 8 (parse-partial-sexp (defun-beginning-position) (point))))))
	 (when pos
	   (goto-char pos)
	   (setq th-string-beginningp t)))))

(put 'string 'end-op-at
     (lambda ()
       (when th-string-beginningp
       (let ((pos (progn (save-excursion (beginning-of-defun) (point))))) 
	 (forward-char 1) 
	 (while (not (eq (char-after) (nth 3 (with-syntax-table (standard-syntax-table) (parse-partial-sexp pos (point))))))
	   (forward-char 1)))
       (forward-char 1)))) 

;; Sh-struct
(put 'sh-struct 'beginning-op-at 'sh-goto-opening)
(put 'sh-struct 'end-op-at
     (lambda ()
       (sh-goto-closing)
       (forward-char 1))) 
	     
(defcustom beg-sh-struct-atpt "case\\|for\\|function\\|if\\|select\\|until\\|while"
  "Specify the sh-struct beginning."
  :type 'regexp
  :group 'convenience)

(defcustom end-sh-struct-atpt "done\\|esac\\|fi" 
  "Specify the sh-struct end."
  :type 'regexp
  :group 'convenience)

;; Symbol

(put 'symbol 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-syntax-backward "w_"))))

(put 'symbol 'end-op-at
     (lambda ()
       (skip-syntax-forward "w_")))

;; Dollared

(put 'dollared 'beginning-op-at
     (lambda ()
       (if (looking-at "$")
	   (point)
	 (search-backward "$" nil nil 1))))

(put 'dollared 'end-op-at
     (lambda ()
       (forward-char 1) 
       (search-forward "$" nil nil 1)))

;; Url

;; derived from thingatpt.el
(defvar thing-at-point-utils-uri-schemes
  ;; Officials from http://www.iana.org/assignments/uri-schemes.html
  '("ftp" "http" "gopher" "mailto" "news" "nntp"
    "telnet" "wais" "file" "prospero" "z39" 
    "cid" "mid" "vemmi" "service" "imap" "nfs" "acap" "rtsp"
    "tip" "pop" "data" "dav" "opaquelocktoken" "sip" "tel" "fax"
    "modem" "ldap" "https" "soap" "urn" "go"
    "afs" "tn3270" "mailserver"
    "crid" "dict" "dns" "dtn" "h323" "im" "info" "ipp"
    "iris" "mtqp" "mupdate" "pres" "sips" "snmp" "tag"
    "tftp" "xmlrpc" "xmpp"
  ;; Compatibility
    "snews" "irc" "mms" "mmsh")
  "Uniform Resource Identifier (URI) Schemes.")

(put 'url 'beginning-op-at
     (lambda ()
       ;; provide for the case, we are over a
       ;; string-delimiter as `"'
       (when
	   (and (not (eq 32 (char-after))) 
		(or (bobp)
		    (eq 32 (char-before))))
	 (forward-char 1)
	 ;; as the bounds-function checks position, correct it
	 ;; (setq th-orig 1)
         )
       (dolist (protokoll thing-at-point-utils-uri-schemes)
         (if (string= (word-atpt) (format "%s" protokoll))
             (goto-char (beginning-of-word-atpt)))
         (search-backward protokoll (line-beginning-position) t 1))))

(put 'url 'end-op-at
     (lambda ()
       ;; (skip-chars-forward ":/?#[]@!$&()*+,;=[:alnum:]-._~")
       (skip-chars-forward "url-at-point-chars")
	 (skip-chars-backward ":")))
			   
(defcustom url-at-point-chars ":/?#[]@!$&()*+,;=[:alnum:]-._~"
  "Chars which might compose a URL. " 
  :type 'string
  :group 'convenience)


;; Whitespace

(put 'whitespace 'beginning-op-at
     (lambda () (when (looking-at "[ \t]") (skip-chars-backward "[ \t\r\n\f[:blank:]]"))))

(put 'whitespace 'end-op-at (lambda () (skip-chars-forward "[ \t\r\n\f[:blank:]]")))

;; Word

(put 'word 'beginning-op-at
     (lambda () (unless (or (bobp)(looking-back "\\W"))
		  (forward-word -1)))) 

(put 'word 'end-op-at
          (lambda () (forward-word 1))) 

     
(defun list-of-things-atpt ()
  "Displays a list of objects which might be called as THING herewith
  Every THING is provided with a set of functions at, before and after point 
  - i.e. to call with ACTION-THING-at,  -bf, -af."
  (interactive)
  (message "%s" thingatpt-listed))

(defun th (thing &optional arg ispec delimiter) 
  "Returns a buffer substring according to THING.
  THING may be a well known form as `symbol',
  `list', `sexp', `defun'.
  You may also define new and abstract kinds of THING.
  See example given in thingatpt-utils-base.el.
  Further functions with `th' provide moves, transpositions.  
  Call `list-of-things-at' to see what's implemented.
  "
  ;; Special handling of `delimited-at':
  ;; As a wide range of delimiters are feasible,
  ;; specify them on the fly.
  ;; (when (and (eq thing 'delimited) delimiters-at)
  (let ((arg (or arg arg 0))) 
    ;; (when (and (eq thing 'delimited) (listp arg)) 
    ;; (set-delimiter-atpt delimiters-at))
    (when (and (eq thing 'csv)
	       (listp arg))
      (setq csv-separator-atpt (read-from-minibuffer "With C-u 4 specify new csv-separator-atpt: " (if (string= csv-separator-atpt ";")
                                                                                                       (regexp-quote ",")
                                                                                                     (regexp-quote ";"))))
      (setq arg 1))
    (condition-case nil
	(let* ((bounds-point-at (th-bounds thing arg))
	       (type 
		(if (and (car bounds-point-at)(cdr bounds-point-at))
		    (buffer-substring-no-properties (car bounds-point-at) (cdr bounds-point-at))
		  nil)))
	  (when ispec
	    (kill-new type)
	    (message "   %s" type))
	  type)
      (error nil))))

(defun th-bounds (thing &optional arg ispec)
  (save-excursion 
    (condition-case nil
	(let* ((arg (or arg arg 0)) 
               (bounds
                (progn
                  (when (< 0 (abs arg))
                    (funcall (intern-soft (concat (symbol-name thing)"-forward-atpt")) arg))
                  (th-bounds-intern thing))))
	  (when ispec (message "%s" bounds))
	  bounds)
      (error nil))))

(defun th-bounds-intern (thing &optional arg)
  "Determine the start and end buffer locations for the THING at point.
  THING is a symbol which specifies the kind of syntactic entity you want.
  Possibilities include `symbol', `list', `sexp', `defun', `filename', `url',
  `word', `sentence', `whitespace', `line', `page' and others.
  Call THING by his name, i.e. word-atpt etc., see `list-of-things-at' to see what's implemented"
  (condition-case nil
      (save-excursion
	(let ((orig (point)) 
	      (beg (progn
		     (funcall (get thing 'beginning-op-at))
		     (point)))
	      (end (progn
		     (funcall (get thing 'end-op-at))
		     (point))))
	  ;; with (< orig end) forward-forms, which end
	  ;; after thing, are usable
	  (when (and (<= orig end)(<= beg orig))
	    (cons beg end))))
    (error nil)))

(defun th-beg (thing arg &optional ispec)
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
	     (beg (car bounds)))
	(when ispec
	  (message "   %s " beg)) 
	beg)
    (error nil)))

(defun th-end (thing arg &optional ispec)
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
	     (end (cdr bounds)))
	(when ispec
	  (message "   %s "  end)) 
	end)
    (error nil)))

(defun th-gotobeg (thing arg &optional ispec)
  "Goto char beginning, core function "
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
	     (beg (car bounds)))
	(when ispec
	  (message "   %s " beg))
	(goto-char beg))
    (error nil)))

(defun th-gotoend (thing arg &optional ispec)
  "Goto char end, core function "
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
	     (end (cdr bounds)))
	(when ispec
	  (message "   %s " end))
	(goto-char end))
    (error nil)))

(defun th-length (thing arg &optional ispec)
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
	     (length (abs (- (car bounds) (cdr bounds)))))
	(when ispec
	  (message "   %s " length)) 
	length)
    (error nil)))

(defun th-escape (thing arg)
  " "
  (regexp-quote (buffer-substring-no-properties (th-beg thing arg)(th-end thing arg))))

(defun th-check (beg end arg &optional ispec)
  "Count letters, number etc. at point. Relate them."
  (let*
      ((refchars
        (if
            (string-match "[/-]" (buffer-substring-no-properties beg end))
            t
          nil))
       (dots
        (if
            (string-match "[\.]" (buffer-substring-no-properties beg end))
            t
          nil))
       (letters (count-matches "[a-z-A-ZäöüÄÖÜß]" beg end))
       (numbers (count-matches "[0-9]" beg end))
       (printables (count-matches "[[:print:]]" beg end))
       (alph-num-ratio
        (when (< 0 numbers)
          (/ letters numbers)))
       (alph-num-rest
        (when (< 0 numbers)
          (% letters numbers))))
    (when ispec
      (message "letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end \n %s %s %s %s %s %s %s %s %s" letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end))
    (list letters numbers printables alph-num-ratio alph-num-rest refchars dots beg end)))

(defun th-copy (thing arg &optional ispec)
  (condition-case nil 
      (let ((newcopy (th thing arg)))
	(if newcopy
	    (progn
	      (kill-new newcopy)
	      (if ispec
		  (message "   %s" (car kill-ring))
		(car kill-ring)))
	  nil))
    (error nil)))

(defun th-trim (&optional thing left right)
  (let* ((bounds (th-bounds thing))
         (left (when left (car bounds)))
         (right (when right (cdr bounds))))
    (save-excursion
      (cond ((and left right)
             (goto-char right)
             (delete-char -1)
             (goto-char left)
             (delete-char 1))
            (right
             (goto-char right)
             (delete-char -1))
	  (left
	   (goto-char left)
	   (delete-char 1))))))

(defun th-separate (thing arg &optional ispec) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg))
             (beg (car bounds))
             (end (if (listp (cdr bounds))
                      (cadr bounds)
                    (cdr bounds)))
             (oldbufsize (buffer-size)))
        (goto-char beg)
        (when (and beg end (<= (line-beginning-position) beg) (not (looking-back "^[ \t]+")))
          (goto-char beg)
          (newline-and-indent))
        (save-excursion 
          (when
              (< oldbufsize (buffer-size)) 
            (setq end (+ end (- (buffer-size) oldbufsize)))
            (setq beg (+ beg (- (buffer-size) oldbufsize)))
            (setq oldbufsize (buffer-size)))
          (goto-char end)
          (cond ((eobp)
                 (newline-and-indent))
                ((looking-at "[\t\r\n\f ]*$")
		      nil)
		     (t (newline-and-indent)))
	       (list beg end)))
    (error nil)))

(defun th-comment (thing arg &optional ispec) 
  " "
  (condition-case nil 
      (let* ((bounds (th-separate thing arg ispec))
	     (beg (car bounds))
	     (end (cadr bounds)))
	(if (and beg end)
	    (progn 
	      (goto-char beg)
	      (comment-or-uncomment-region beg (1+ end)))
	  nil))
    (error nil))) 

(defun th-kill (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg)) 
	     (beg (car bounds))
	     (end (cdr bounds)))
	(kill-region beg end))
    (error nil)))

(defun th-mark (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing arg)))
	;; (if (not (region-active-p))
	;; (error "%s" "Region must be active")
	(goto-char (car bounds))
	(push-mark)
	(goto-char (cdr bounds))
	(exchange-point-and-mark))
    (error nil)))

(defun th-forward (thing arg &optional ispec) 
  " "
  (condition-case nil 
      (let* ((orig (point))
	     (arg (if arg (prefix-numeric-value arg) 1))
	     (bounds (th-bounds-intern thing))
	     (ap (car bounds))
	     (ep (cdr bounds)))
	(if (< 0 arg)
	    (progn 
	      (if (and ep (< orig (1- ep)))
		  (progn (goto-char (1- ep))
			 (setq arg (1- arg)))
		(forward-char 1)) 
	      (while (< 0 arg)
		(forward-char 1) 
		;; otherwise containing list is taken
		(if (and (eq thing 'list) (not (eq 4 (syntax-class (syntax-after (point)))))) 
		    (skip-syntax-forward "^("))
                (while (and (not (eobp))(not (setq ep (cdr (th-bounds-intern thing)))))
                  (forward-char 1))
                (setq arg (1- arg))
                (goto-char ep))
              (goto-char (1- ep)))
          ;; (> 0 arg)
          (when (and ap (< ap orig))
            (goto-char ap)
            (setq arg (1+ arg))) 
          (while
              (> 0 arg)
            (forward-char -1)
            ;; otherwise containing list is taken
            (if (and (eq thing 'list) (not (eq 5 (syntax-class (syntax-after (point))))))
                (progn 
                  (skip-syntax-backward "^)")
		  (forward-char -1))
              (while (and (not (bobp)) (not (setq ap (car (th-bounds-intern thing)))))
		  (forward-char -1))
		(setq arg (1+ arg))
		(goto-char ap))))
	(when ispec (message "%s" (point))) 
	(point))
    (error nil)))

(defun th-backward (thing arg &optional ispec) 
  " "
  (condition-case nil
      (th-forward thing (- (or arg 1)))
    (when ispec (message "%s" (point)))
    (point)
    (error nil)))

(defun th-transpose (thing arg)
  (let* ((arg (if arg (prefix-numeric-value arg) 1))
	 (pos (point)) 
	 (pos1 (or (th-bounds thing (- (1+ arg)))))
	 (pos2 (th-bounds thing 0))
	 word1 word2)
    ;; borrowed from simple.el
    (if (featurep 'xemacs)
	(progn (narrow-to-region (car pos1) (cdr pos2))
	       (setq word2 (delete-and-extract-region (car pos2) (cdr pos2)))
	       (goto-char (car pos2))
	       (insert (delete-and-extract-region (car pos1) (cdr pos1)))
	       (goto-char (car pos1))
	       (insert word2)
	       (goto-char (point-max)))
      (atomic-change-group
	(narrow-to-region (car pos1) (cdr pos2))
	(setq word2 (delete-and-extract-region (car pos2) (cdr pos2)))
	(goto-char (car pos2))
	(insert (delete-and-extract-region (car pos1) (cdr pos1)))
	(goto-char (car pos1))
	(insert word2)
	(goto-char (point-max))))
    (widen)))

;; sort-subr, sort.el
;; (reverse nextrecfun endrecfun &optional startkeyfun endkeyfun predicate)

(defun th-sort (thing reverse beg end startkeyfun endkeyfun predicate)
  (if (region-active-p) 
      (save-excursion
	(save-restriction
	  (narrow-to-region beg end)
	  (goto-char (point-min))
	  (let ((reverse (or reverse nil))
		(startkeyfun (or startkeyfun nil))
		(endkeyfun (or endkeyfun nil))
		(predicate (or predicate nil))
		(this-beg (th-beg thing 0)))
	    (unless (and this-beg (eq this-beg (point))))
	    (th-forward thing 1)
	    (th-gotobeg thing 0)
	    (sort-subr reverse (function (lambda () (let ((pos (point))) (th-forward thing 1) (if (< pos (point)) (th-gotobeg thing 0) nil)))) (function (lambda () (th-gotoend thing 1))) startkeyfun endkeyfun predicate))))
    (error "Region must be active!")))

(defun th-delim (thing arg &optional ispec beg-char end-char)
  "Process begin and end of region according to value of 
  `delim-action'
  If no region is active, process borders of THING-at-point
  according to value of delim-action-beginning- resp. -end-position
  Default is symbol-at.
  With \C-u or arg `escaped' to `t': insert escaped doublequotes"
  (interactive "*P\np")
  (condition-case nil 
      (save-restriction
	(let* ((pos (point)) 
	       (bounds (th-bounds thing arg)) 
	       (beg (car bounds))
	       (end (cdr bounds))
	       (beg-char (or beg-char th-beg-delimiter))
	       (end-char (or end-char th-end-delimiter)))
	  (narrow-to-region beg end)
	  (point-min-marker)
	  (point-max-marker)
	  (goto-char beg)
	  (insert beg-char)
	  (goto-char (point-max-marker))
	  (insert end-char)
	  (goto-char (1+ pos))
	  (when ispec (message "%s" bounds))))
    (error nil)))

(defun th-backslash (thing arg &optional ispec) 
  " "
  (th-delim thing arg ispec backlash-beg-char backlash-end-char))

(setq backlash-end-char ?\\)
(setq backlash-beg-char ?\\)

(defun th-brace (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec "\{" "\}"))

(defun th-bracket (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\[ ?\]))

(defun th-dollar (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\$ ?\$))

(defun th-doublequote (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\" ?\"))

(defun th-singlequote (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\' ?\'))

(defun th-parentize (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\( ?\)))

(defun th-slash (thing arg &optional ispec)
  " "
  (th-delim thing arg ispec ?\/ ?\/))

(defun syntax-class-atpt (&optional ispec) 
  ""
  (interactive "p")
  (let ((sc (syntax-atpt)))
    (if ispec
	(message "%s" sc)
      sc)))

(defun syntax-class-bf (&optional ispec) 
  ""
  (interactive "p")
  (let ((sc (syntax-bf)))
    (if ispec
	(message "%s" sc)
      sc)))

(defun syntax-atpt (&optional ispec) 
  (interactive "p")
  (let ((stax (syntax-after (point))))
    (when ispec
      (message (format "%s" stax)))
    (car stax)))

(defun syntax-bf (&optional ispec) 
  (interactive "p")
  (let ((stax (syntax-after (1- (point)))))
    (when ispec
      (message (format "%s" stax)))
    (car stax)))

;; Until then when `up-list' was called while in
;;  string, it failed with message: 
;; Scan error: ... "Unbalanced parentheses" 

(defun list-up-atpt (&optional arg)
  "Move forward out of one level of parentheses.
With ARG, do this that many times.
A negative argument means move backward but still to a less deep spot."
  (interactive "p")
  (let ((inc (if (> arg 0) 1 -1)))
    (while (/= arg 0)
      (goto-char
       (or
	(progn
	  (when (in-string-p)
	    (goto-char (or (nth 8 (parse-partial-sexp (defun-beginning-position) (point))) (point)))
	    (forward-sexp 1))
	  (scan-lists (point) inc 1)) (buffer-end arg)))
      (setq arg (- arg inc)))))

(defun list-forward-atpt (&optional arg)
  "Move forward across one balanced group of parentheses.
  With ARG, do it that many times.
  Negative arg -N means move backward across N groups of parentheses."
  (interactive "p")
  (let ((pos (point))
	(uplist-arg (if (< 0 arg)
			1
		      -1))
	(akt-open (nth 1 (parse-partial-sexp 1 (point)))))
    (if (eq 4 (syntax-class (syntax-after (point))))
	(forward-list arg)
      (if (in-string-p)
	  (progn 
	    (goto-char (or (nth 8 (parse-partial-sexp 1 (point))) (point)))
	    (forward-char -1)
	    (up-list uplist-arg)
	    (if (< 0 arg)
		(setq arg (1- arg)) 
	      (setq arg (1+ arg))))
	(when (and (< 0 (abs arg)) (list-atpt arg))
	  (up-list uplist-arg)
	  (if (< 0 arg)
	      (setq arg (1- arg))
	    (setq arg (1+ arg)))))
      (forward-list arg))))

(defun delist-list-atpt (arg)
  "Delete delimiters of a list at point"
  (interactive "*p")
  (let ((poss (list-bounds-atpt arg)))
    (narrow-to-region (car poss) (if (listp (cdr poss))
				     (cadr poss)
				   (cdr poss)))
    (point-min-marker)
    (point-max-marker) 
    (goto-char (point-min-marker))
    (delete-char 1)
    (goto-char (point-max-marker)) 
    (backward-delete-char 1)
    (widen)))


(provide 'thingatpt-utils-base)
;;; thingatpt-utils-base.el ends here
